'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

exports.wareBuilder = wareBuilder;

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

function flatten(arg) {
  return arg.reduce(function (memo, item) {
    return Array.isArray(item) ? [].concat(_toConsumableArray(memo), _toConsumableArray(flatten(item))) : [].concat(_toConsumableArray(memo), [item]);
  }, []);
}

function normalizeArgumentArray(args) {
  return flatten(args).filter(function (item) {
    return !!item;
  });
}

function isFunction(fn) {
  return typeof fn === 'function';
}

function isPromise(obj) {
  return !!obj && ((typeof obj === 'undefined' ? 'undefined' : _typeof(obj)) === 'object' || typeof obj === 'function') && typeof obj.then === 'function';
}

var DEFAULT_COMPOSER = function DEFAULT_COMPOSER(fn) {
  return fn;
};
var DEFAULT_IS_ERROR_METHOD = function DEFAULT_IS_ERROR_METHOD(fn, arity) {
  return fn.length === arity + 1;
};

function wareBuilder() {
  var composer = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : DEFAULT_COMPOSER;
  var errorComposer = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : composer;

  var _ref = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {},
      _ref$isErrorMethod = _ref.isErrorMethod,
      isErrorMethod = _ref$isErrorMethod === undefined ? DEFAULT_IS_ERROR_METHOD : _ref$isErrorMethod;

  // high level - returns a function that can merge results
  return function middle() {
    for (var _len = arguments.length, middleware = Array(_len), _key = 0; _key < _len; _key++) {
      middleware[_key] = arguments[_key];
    }

    // this is the actual warewolf() call
    return function () {
      for (var _len2 = arguments.length, invocationArguments = Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
        invocationArguments[_key2] = arguments[_key2];
      }

      var hasCallback = true;
      var middlewareQueue = normalizeArgumentArray(middleware);
      var isDone = false;
      // some assumptions here about how warewolf is called
      // if callback is provided, that's subbed with next() in middleware
      // if no callback, we treat it like a promise
      var arity = invocationArguments.length;

      var done = invocationArguments.pop();
      if (!isFunction(done)) {
        if (done) {
          invocationArguments.push(done);
        }
        hasCallback = false;
      }

      var promiseBody = function promiseBody(resolve, reject) {
        if (!hasCallback) {
          done = function done(err, result) {
            if (err) {
              reject(err);
            } else {
              resolve(result);
            }
          };
        }

        function moveToNext(err) {
          while (middlewareQueue.length && (err ? !isErrorMethod(middlewareQueue[0], arity, invocationArguments) : isErrorMethod(middlewareQueue[0], arity, invocationArguments))) {
            middlewareQueue.shift();
          }
          if (middlewareQueue.length) {
            var fn = middlewareQueue.shift();
            return err ? errorComposer(fn.bind(this, err)) : composer(fn);
          }
          return null;
        }

        function nextStep(index, err) {
          if (index !== middlewareQueue.length) {
            isDone = true;
            done('Error: next() has been called multiple times. Be careful not to return a promise and call a callback in the same method.');
            return;
          }
          var nextFn = moveToNext(err);
          var nextStepWithIndex = nextStep.bind(this, middlewareQueue.length);
          try {
            if (!nextFn) {
              isDone = true;

              for (var _len3 = arguments.length, resultFromStep = Array(_len3 > 2 ? _len3 - 2 : 0), _key3 = 2; _key3 < _len3; _key3++) {
                resultFromStep[_key3 - 2] = arguments[_key3];
              }

              done.apply(undefined, [err].concat(resultFromStep));
              return;
            }
            var flowArgs = hasCallback ? [nextStepWithIndex] : [];

            var result = nextFn.apply(undefined, invocationArguments.concat(flowArgs));
            if (isPromise(result)) {
              result.then(function (r) {
                return nextStepWithIndex(null, r);
              }).catch(nextStepWithIndex);
            } else if (!hasCallback) {
              nextStepWithIndex(err, result);
            }
          } catch (e) {
            if (e.name === 'AssertionError') {
              throw e;
            }
            if (middlewareQueue.find(function (fn) {
              return isErrorMethod(fn, arity, invocationArguments);
            })) {
              nextStepWithIndex(e);
              return;
            }
            if (!isDone) {
              done(e);
              return;
            }
            throw e;
          }
        }
        nextStep(middlewareQueue.length);
      };

      if (!hasCallback) {
        return new Promise(promiseBody);
      }
      return promiseBody();
    };
  };
}

exports.default = wareBuilder();